fn main() {
    println!("Hello, world!");
}

mod rover {
    #[derive(Debug, PartialEq, Copy, Clone)]
    pub enum Facing { N, E, S, W }
    #[derive(Debug, PartialEq, Copy, Clone)]
    pub enum Rotation { Left, Right }

    pub type Rover = (i32, i32, Facing);

    fn velocity(direction : &Facing) -> (i32, i32) {
	match direction {
	    Facing::N  => ( 0, -1),
	    Facing::E  => ( 1,  0),
	    Facing::S  => ( 0,  1),
	    Facing::W  => (-1,  0)
	}
    }

    pub fn forward((lon, lat, direction) : &Rover) -> Rover {
	let (xoff, yoff) = velocity(&direction);
	(lon + xoff as i32, lat + yoff as i32, *direction)
    }

    pub fn backward((lon, lat, direction) : &Rover) -> Rover {
	let (xoff, yoff) = velocity(&direction);
	(lon - xoff as i32, lat - yoff as i32, *direction)
    }

    fn next_direction(direction : Facing) -> Facing {
	match direction {
	    Facing::N => Facing::W,
	    Facing::W => Facing::S,
	    Facing::S => Facing::E,
	    Facing::E => Facing::N,
	}
    }

    pub fn rotate((lon, lat, direction) : &Rover, rotation: Rotation) -> Rover {
	match rotation {
	    Rotation::Left => {
		let new_d = next_direction(*direction);
		(*lon, *lat, new_d)
	    },
	    Rotation::Right => {
		let new_d = next_direction(*direction);
		let new_d = next_direction(new_d);
		let new_d = next_direction(new_d);
		(*lon, *lat, new_d)
	    }
	}
    }
}


mod driver {
    use super::rover;

    fn perform_insn(command: &str, rover: &rover::Rover) -> rover::Rover {
	match command {
	    "f" => rover::forward(&rover),
	    "b" => rover::backward(&rover),
	    "l" => rover::rotate(&rover, rover::Rotation::Left),
	    "r" => rover::rotate(&rover, rover::Rotation::Right),
	    _ => rover.clone()
	}
    }

    pub fn execute(rover: &rover::Rover, instructions: &str) -> rover::Rover {
	if instructions.is_empty() {
	    rover.clone()
	} else {
	    let (first, rest) = instructions.split_at(1);
	    execute(&perform_insn(first, rover), rest)
	}
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn move_north() {
	let r = (1 as i32, 1 as i32, rover::Facing::N);
	assert_eq!(rover::forward(&r),
		   (1 as i32, 0 as i32, rover::Facing::N));

	let r = (1 as i32, 3 as i32, rover::Facing::N);
	assert_eq!(rover::forward(&r),
		   (1 as i32, 2 as i32, rover::Facing::N));
    }

    #[test]
    fn move_south() {
	let r = (1 as i32, 1 as i32, rover::Facing::S);
	assert_eq!(rover::forward(&r),
		   (1 as i32, 2 as i32, rover::Facing::S));
    }

    #[test]
    fn move_west() {
	let r = (1 as i32, 1 as i32, rover::Facing::W);
	assert_eq!(rover::forward(&r),
		   (0 as i32, 1 as i32, rover::Facing::W));
    }

    #[test]
    fn move_east() {
	let r = (1 as i32, 1 as i32, rover::Facing::E);
	assert_eq!(rover::forward(&r),
		   (2 as i32, 1 as i32, rover::Facing::E));
    }

    #[test]
    fn move_backward() {
	let r = (1 as i32, 1 as i32, rover::Facing::E);
	let r2 = rover::forward(&r);
	assert_eq!(rover::backward(&r2), r);
    }

    #[test]
    fn spin_left() {
	let r = (1 as i32, 1 as i32, rover::Facing::E);
	match rover::rotate(&r, rover::Rotation::Left) {
	    (_, _, facing) => assert_eq!(facing, rover::Facing::N)
	}
    }
    #[test]
    fn spin_right() {
	let r = (1 as i32, 1 as i32, rover::Facing::S);
	match rover::rotate(&r, rover::Rotation::Right) {
	    (_, _, facing) => assert_eq!(facing, rover::Facing::W)
	}
    }

    #[test]
    fn drive() {
	let r = (1 as i32, 1 as i32, rover::Facing::S);
	assert_eq!(driver::execute(&r, "ffrf"),
		   (0 as i32, 3 as i32, rover::Facing::W));
    }

}
